/*
 * QEMU model of the BBRAM Battery Backed RAM
 *
 * Copyright (c) 2014 Xilinx Inc.
 *
 * Autogenerated by xregqemu.py 2014-05-14.
 *
 * Permission is hereby granted, free of charge, to any person obtaining a copy
 * of this software and associated documentation files (the "Software"), to deal
 * in the Software without restriction, including without limitation the rights
 * to use, copy, modify, merge, publish, distribute, sublicense, and/or sell
 * copies of the Software, and to permit persons to whom the Software is
 * furnished to do so, subject to the following conditions:
 *
 * The above copyright notice and this permission notice shall be included in
 * all copies or substantial portions of the Software.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
 * IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
 * FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
 * THE AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
 * LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM,
 * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN
 * THE SOFTWARE.
 */

#include "qemu/osdep.h"
#include "hw/sysbus.h"
#include "hw/register-dep.h"
#include "qemu/error-report.h"
#include "qemu/bitops.h"
#include "qemu/log.h"
#include "qapi/error.h"
#include "sysemu/blockdev.h"
#include "sysemu/block-backend.h"

#include "hw/zynqmp_aes_key.h"

#ifndef XILINX_BBRAM_CTRL_ERR_DEBUG
#define XILINX_BBRAM_CTRL_ERR_DEBUG 0
#endif

#define TYPE_XILINX_BBRAM_CTRL "xlnx.bbram-ctrl"

#define XILINX_BBRAM_CTRL(obj) \
     OBJECT_CHECK(BBRAMCtrl, (obj), TYPE_XILINX_BBRAM_CTRL)

#define DB_PRINT_L(lvl, fmt, args...) do { \
    if (XILINX_BBRAM_CTRL_ERR_DEBUG >= lvl) { \
        qemu_log("%s: " fmt, __func__, ## args); \
    } \
} while (0);

#define DB_PRINT(fmt, args...) DB_PRINT_L(1, fmt, ## args)

DEP_REG32(BBRAM_STATUS, 0x0)
    DEP_FIELD(BBRAM_STATUS, AES_CRC_PASS, 1, 9)
    DEP_FIELD(BBRAM_STATUS, AES_CRC_DONE, 1, 8)
    DEP_FIELD(BBRAM_STATUS, BBRAM_ZEROIZED, 1, 4)
    DEP_FIELD(BBRAM_STATUS, PGM_MODE, 1, 0)
DEP_REG32(BBRAM_CTRL, 0x4)
    DEP_FIELD(BBRAM_CTRL, ZEROIZE, 1, 0)
DEP_REG32(PGM_MODE, 0x8)
DEP_REG32(BBRAM_AES_CRC, 0xc)
DEP_REG32(BBRAM_0, 0x10)
DEP_REG32(BBRAM_1, 0x14)
DEP_REG32(BBRAM_2, 0x18)
DEP_REG32(BBRAM_3, 0x1c)
DEP_REG32(BBRAM_4, 0x20)
DEP_REG32(BBRAM_5, 0x24)
DEP_REG32(BBRAM_6, 0x28)
DEP_REG32(BBRAM_7, 0x2c)
DEP_REG32(BBRAM_8, 0x30)
DEP_REG32(BBRAM_SLVERR, 0x34)
    DEP_FIELD(BBRAM_SLVERR, ENABLE, 1, 0)
DEP_REG32(BBRAM_ISR, 0x38)
    DEP_FIELD(BBRAM_ISR, APB_SLVERR, 1, 0)
DEP_REG32(BBRAM_IMR, 0x3c)
    DEP_FIELD(BBRAM_IMR, APB_SLVERR, 1, 0)
DEP_REG32(BBRAM_IER, 0x40)
    DEP_FIELD(BBRAM_IER, APB_SLVERR, 1, 0)
DEP_REG32(BBRAM_IDR, 0x44)
    DEP_FIELD(BBRAM_IDR, APB_SLVERR, 1, 0)
DEP_REG32(BBRAM_ECO, 0x48)
DEP_REG32(SLV_ERR_TRIG, 0x54)
    DEP_FIELD(SLV_ERR_TRIG, VAL, 1, 0)
DEP_REG32(BBRAM_MSW_LOCK, 0x64)
    DEP_FIELD(BBRAM_MSW_LOCK, VAL, 1, 0)

#define R_MAX (R_BBRAM_MSW_LOCK + 1)

#define ZYNQMP_BBRAM_SIZE (8 * 4)
#define ZYNQMP_PGM_MAGIC_MASK 0x757bdf0d
#define IS_ZYNQMP (s->zynqmp_keysink)

#define ZYNQ3_PGM_MAGIC 0x757BDF0F

typedef struct BBRAMCtrl {
    SysBusDevice parent_obj;
    MemoryRegion iomem;
    qemu_irq irq_bbram;

    BlockBackend *blk;
    ZynqMPAESKeySink *zynqmp_keysink;
    uint32_t *ram32;
    uint32_t size;
    bool msw_lock;

    uint32_t regs[R_MAX];
    DepRegisterInfo regs_info[R_MAX];
} BBRAMCtrl;

static void bbram_ram_sync(BBRAMCtrl *s)
{
    /* Check if there is a ZynqMP key */
    if (IS_ZYNQMP) {
        if (0) {
            /* Real HW does only update the AES key at power on reset. */
            zynqmp_aes_key_update(s->zynqmp_keysink,
                                  (void *) &s->regs[R_BBRAM_0],
                                  ZYNQMP_BBRAM_SIZE);
        }
        return;
    }

    if (!s->blk) {
        return;
    }

    memcpy(s->ram32, &s->regs[R_BBRAM_0], (R_BBRAM_8 - R_BBRAM_0) * 4);
    if (blk_pwrite(s->blk, 0, (void *) s->ram32, s->size, 0) < 0) {
        error_report("%s: write error in sector", __func__);
    }
}

static void bbram_zeroize(BBRAMCtrl *s)
{
    DB_PRINT("Zeroing out the key\n");
    memset(&s->regs[R_BBRAM_0], 0, (R_BBRAM_8 - R_BBRAM_0) * 4);
    bbram_ram_sync(s);
}

static void bbram_update_irq(BBRAMCtrl *s)
{
    bool pending = s->regs[R_BBRAM_ISR] & ~s->regs[R_BBRAM_IMR];

    DB_PRINT("Setting the interrupt: %d\n", pending);
    qemu_set_irq(s->irq_bbram, pending);
}

static void bbram_ctrl_postw(DepRegisterInfo *reg, uint64_t val64)
{
    BBRAMCtrl *s = XILINX_BBRAM_CTRL(reg->opaque);
    uint32_t val = val64;

    if (val & R_BBRAM_CTRL_ZEROIZE_MASK) {
        bbram_zeroize(s);
        s->regs[R_BBRAM_STATUS] |= R_BBRAM_STATUS_BBRAM_ZEROIZED_MASK;
        /* The bit is self clearing */
        s->regs[R_BBRAM_CTRL] &= ~R_BBRAM_CTRL_ZEROIZE_MASK;
    }
}

static void bbram_pgm_mode_postw(DepRegisterInfo *reg, uint64_t val64)
{
    BBRAMCtrl *s = XILINX_BBRAM_CTRL(reg->opaque);
    uint32_t val = val64;

    if (val == ZYNQMP_PGM_MAGIC_MASK) {
        bbram_zeroize(s);
        s->regs[R_BBRAM_STATUS] |= R_BBRAM_STATUS_PGM_MODE_MASK;
    } else if (val == ZYNQ3_PGM_MAGIC) {
        bbram_zeroize(s);
        s->regs[R_BBRAM_STATUS] |= R_BBRAM_STATUS_PGM_MODE_MASK;
    }
}

static void bbram_aes_crc_postw(DepRegisterInfo *reg, uint64_t val64)
{
    BBRAMCtrl *s = XILINX_BBRAM_CTRL(reg->opaque);

    if (!(s->regs[R_BBRAM_STATUS] & R_BBRAM_STATUS_PGM_MODE_MASK)) {
        /* We are not in programming mode, don't do anything */
        return;
    }

    /* Perform the AES integrity check */
    s->regs[R_BBRAM_STATUS] |= R_BBRAM_STATUS_AES_CRC_DONE_MASK;

    /* If the value calculated by the BBRAM controller matches the value
     * written to this register the test passes. We don't know what the test
     * is so let's just pass it.
     */
    s->regs[R_BBRAM_STATUS] |= R_BBRAM_STATUS_AES_CRC_PASS_MASK;
}

static uint64_t bbram_key_prew(DepRegisterInfo *reg, uint64_t val64)
{
    BBRAMCtrl *s = XILINX_BBRAM_CTRL(reg->opaque);
    uint32_t original_data = *(uint32_t *) reg->data;

    if (s->regs[R_BBRAM_STATUS] & R_BBRAM_STATUS_PGM_MODE_MASK) {
        if (s->msw_lock) {
            DB_PRINT("MSW lock is set. Writing value: 0x%"HWADDR_PRIx"\n",
                     val64 & 0xFFFF);
            return val64 & 0xFFFF;
        } else {
            DB_PRINT("Writing value: 0x%"HWADDR_PRIx"\n", val64);
            return val64;
        }
    } else {
        /* We are not in programming mode, don't do anything */
        qemu_log_mask(LOG_GUEST_ERROR,
                      "Not in programming mode, dropping the write\n");
        return original_data;
    }
}

static void bbram_key_postw(DepRegisterInfo *reg, uint64_t val64)
{
    BBRAMCtrl *s = XILINX_BBRAM_CTRL(reg->opaque);

    if (IS_ZYNQMP) {
        zynqmp_aes_key_update(s->zynqmp_keysink, (void *) &s->regs[R_BBRAM_0],
                              ZYNQMP_BBRAM_SIZE);
    }
}

static void bbram_isr_postw(DepRegisterInfo *reg, uint64_t val64)
{
    BBRAMCtrl *s = XILINX_BBRAM_CTRL(reg->opaque);

    bbram_update_irq(s);
}

static uint64_t bbram_ier_prew(DepRegisterInfo *reg, uint64_t val64)
{
    BBRAMCtrl *s = XILINX_BBRAM_CTRL(reg->opaque);
    uint32_t val = val64;

    s->regs[R_BBRAM_IMR] &= ~val;
    bbram_update_irq(s);
    return 0;
}

static uint64_t bbram_idr_prew(DepRegisterInfo *reg, uint64_t val64)
{
    BBRAMCtrl *s = XILINX_BBRAM_CTRL(reg->opaque);
    uint32_t val = val64;

    s->regs[R_BBRAM_IMR] |= val;
    bbram_update_irq(s);
    return 0;
}

static uint64_t slv_err_trig_prew(DepRegisterInfo *reg, uint64_t val64)
{
    BBRAMCtrl *s = XILINX_BBRAM_CTRL(reg->opaque);
    uint32_t val = val64;

    s->regs[R_BBRAM_ISR] |= val;
    bbram_update_irq(s);
    return 0;
}

static uint64_t bbram_msw_lock_prew(DepRegisterInfo *reg, uint64_t val64)
{
    BBRAMCtrl *s = XILINX_BBRAM_CTRL(reg->opaque);
    uint32_t val = val64;

    if (s->msw_lock) {
        DB_PRINT("MSW lock is enanabled and can't be changed\n");
        return 1 << R_BBRAM_MSW_LOCK_VAL_SHIFT;
    } else if (val & (R_BBRAM_MSW_LOCK_VAL_MASK)) {
        s->msw_lock = true;
    }

    return 0;
}

static DepRegisterAccessInfo bbram_ctrl_regs_info[] = {
    {   .name = "BBRAM_STATUS",  .decode.addr = A_BBRAM_STATUS,
        .rsvd = 0xee,
        .ro = 0x3ff,
    },{ .name = "BBRAM_CTRL",  .decode.addr = A_BBRAM_CTRL,
        .post_write = bbram_ctrl_postw,
    },{ .name = "PGM_MODE",  .decode.addr = A_PGM_MODE,
        .post_write = bbram_pgm_mode_postw,
    },{ .name = "BBRAM_AES_CRC",  .decode.addr = A_BBRAM_AES_CRC,
        .post_write = bbram_aes_crc_postw,
    },{ .name = "BBRAM_0",  .decode.addr = A_BBRAM_0,
        .pre_write = bbram_key_prew,
        .post_write = bbram_key_postw,
    },{ .name = "BBRAM_1",  .decode.addr = A_BBRAM_1,
        .pre_write = bbram_key_prew,
        .post_write = bbram_key_postw,
    },{ .name = "BBRAM_2",  .decode.addr = A_BBRAM_2,
        .pre_write = bbram_key_prew,
        .post_write = bbram_key_postw,
    },{ .name = "BBRAM_3",  .decode.addr = A_BBRAM_3,
        .pre_write = bbram_key_prew,
        .post_write = bbram_key_postw,
    },{ .name = "BBRAM_4",  .decode.addr = A_BBRAM_4,
        .pre_write = bbram_key_prew,
        .post_write = bbram_key_postw,
    },{ .name = "BBRAM_5",  .decode.addr = A_BBRAM_5,
        .pre_write = bbram_key_prew,
        .post_write = bbram_key_postw,
    },{ .name = "BBRAM_6",  .decode.addr = A_BBRAM_6,
        .pre_write = bbram_key_prew,
        .post_write = bbram_key_postw,
    },{ .name = "BBRAM_7",  .decode.addr = A_BBRAM_7,
        .pre_write = bbram_key_prew,
        .post_write = bbram_key_postw,
    },{ .name = "BBRAM_8",  .decode.addr = A_BBRAM_8,
        .pre_write = bbram_key_prew,
        .post_write = bbram_key_postw,
    },{ .name = "BBRAM_SLVERR",  .decode.addr = A_BBRAM_SLVERR,
    },{ .name = "BBRAM_ISR",  .decode.addr = A_BBRAM_ISR,
        .w1c = 0x1,
        .post_write = bbram_isr_postw,
    },{ .name = "BBRAM_IMR",  .decode.addr = A_BBRAM_IMR,
        .ro = 0x1,
    },{ .name = "BBRAM_IER",  .decode.addr = A_BBRAM_IER,
        .pre_write = bbram_ier_prew,
    },{ .name = "BBRAM_IDR",  .decode.addr = A_BBRAM_IDR,
        .pre_write = bbram_idr_prew,
    },{ .name = "SLV_ERR_TRIG",  .decode.addr = A_SLV_ERR_TRIG,
        .pre_write = slv_err_trig_prew,
    },{ .name = "BBRAM_ECO",  .decode.addr = A_BBRAM_ECO,
    },{ .name = "BBRAM_MSW_LOCK",  .decode.addr = A_BBRAM_MSW_LOCK,
        .pre_write = bbram_msw_lock_prew,
    }
};

static void bbram_ctrl_reset(DeviceState *dev)
{
    BBRAMCtrl *s = XILINX_BBRAM_CTRL(dev);
    unsigned int i;

    for (i = 0; i < ARRAY_SIZE(s->regs_info); ++i) {
        if (i < R_BBRAM_0 || i > R_BBRAM_8) {
            dep_register_reset(&s->regs_info[i]);
        }
    }

    if (IS_ZYNQMP) {
        zynqmp_aes_key_update(s->zynqmp_keysink, (void *) &s->regs[R_BBRAM_0],
                              ZYNQMP_BBRAM_SIZE);
    }

    bbram_update_irq(s);
}

static uint64_t bbram_ctrl_read(void *opaque, hwaddr addr, unsigned size)
{
    BBRAMCtrl *s = XILINX_BBRAM_CTRL(opaque);
    DepRegisterInfo *r = &s->regs_info[addr / 4];

    if (!r->data) {
        qemu_log("%s: Decode error: read from %" HWADDR_PRIx "\n",
                 object_get_canonical_path(OBJECT(s)),
                 addr);
        return 0;
    }
    return dep_register_read(r);
}

static void bbram_ctrl_write(void *opaque, hwaddr addr, uint64_t value,
                      unsigned size)
{
    BBRAMCtrl *s = XILINX_BBRAM_CTRL(opaque);
    DepRegisterInfo *r = &s->regs_info[addr / 4];

    if (!r->data) {
        qemu_log("%s: Decode error: write to %" HWADDR_PRIx "=%" PRIx64 "\n",
                 object_get_canonical_path(OBJECT(s)),
                 addr, value);
        return;
    }
    dep_register_write(r, value, ~0);
}

static const MemoryRegionOps bbram_ctrl_ops = {
    .read = bbram_ctrl_read,
    .write = bbram_ctrl_write,
    .endianness = DEVICE_LITTLE_ENDIAN,
    .valid = {
        .min_access_size = 4,
        .max_access_size = 4,
    },
};

static void bbram_ctrl_realize(DeviceState *dev, Error **errp)
{
    BBRAMCtrl *s = XILINX_BBRAM_CTRL(dev);
    DriveInfo *dinfo;
    BlockBackend *blk;
    const char *prefix = object_get_canonical_path(OBJECT(dev));
    unsigned int i;

    if (!s->zynqmp_keysink) {
        error_setg(&error_abort,
                   "%s: AES BBRAM key sink not connected\n", prefix);
    }

    dinfo = drive_get_next(IF_PFLASH);
    blk = dinfo ? blk_by_legacy_dinfo(dinfo) : NULL;
    s->ram32 = g_malloc0(s->size);
    memset(s->ram32, 0, s->size);
    if (blk) {
        qdev_prop_set_drive(dev, "drive", blk, errp);
        if (!blk_pread(s->blk, 0, (void *) s->ram32, s->size)) {
            error_setg(&error_abort, "%s: Unable to read-out contents."
                         "backing file too small? Expecting %u bytes",
                          prefix, s->size);
        }
    }
    memcpy(&s->regs[R_BBRAM_0], s->ram32, (R_BBRAM_8 - R_BBRAM_0) * 4);

    for (i = 0; i < ARRAY_SIZE(bbram_ctrl_regs_info); ++i) {
        DepRegisterInfo *r;

        r = &s->regs_info[bbram_ctrl_regs_info[i].decode.addr / 4];
        *r = (DepRegisterInfo) {
            .data = (uint8_t *)&s->regs[
                    bbram_ctrl_regs_info[i].decode.addr / 4],
            .data_size = sizeof(uint32_t),
            .access = &bbram_ctrl_regs_info[i],
            .debug = XILINX_BBRAM_CTRL_ERR_DEBUG,
            .prefix = prefix,
            .opaque = s,
        };
    }
}

static void bbram_ctrl_init(Object *obj)
{
    BBRAMCtrl *s = XILINX_BBRAM_CTRL(obj);
    SysBusDevice *sbd = SYS_BUS_DEVICE(obj);

    memory_region_init_io(&s->iomem, obj, &bbram_ctrl_ops, s,
                          TYPE_XILINX_BBRAM_CTRL, R_MAX * 4);
    sysbus_init_mmio(sbd, &s->iomem);
    sysbus_init_irq(sbd, &s->irq_bbram);
}

static const VMStateDescription vmstate_bbram_ctrl = {
    .name = TYPE_XILINX_BBRAM_CTRL,
    .version_id = 1,
    .minimum_version_id = 1,
    .fields = (VMStateField[]) {
        VMSTATE_UINT32_ARRAY(regs, BBRAMCtrl, R_MAX),
        VMSTATE_END_OF_LIST(),
    }
};

static Property bbram_ctrl_props[] = {
    DEFINE_PROP_LINK("zynqmp-aes-key-sink-bbram", BBRAMCtrl, zynqmp_keysink,
                     TYPE_ZYNQMP_AES_KEY_SINK, ZynqMPAESKeySink *),
    DEFINE_PROP_UINT32("bbram-size", BBRAMCtrl, size, ZYNQMP_BBRAM_SIZE),
    DEFINE_PROP_DRIVE("drive", BBRAMCtrl, blk),
    DEFINE_PROP_END_OF_LIST(),
};

static void bbram_ctrl_class_init(ObjectClass *klass, void *data)
{
    DeviceClass *dc = DEVICE_CLASS(klass);

    dc->reset = bbram_ctrl_reset;
    dc->realize = bbram_ctrl_realize;
    dc->vmsd = &vmstate_bbram_ctrl;
    dc->props = bbram_ctrl_props;
}

static const TypeInfo bbram_ctrl_info = {
    .name          = TYPE_XILINX_BBRAM_CTRL,
    .parent        = TYPE_SYS_BUS_DEVICE,
    .instance_size = sizeof(BBRAMCtrl),
    .class_init    = bbram_ctrl_class_init,
    .instance_init = bbram_ctrl_init,
};

static void bbram_ctrl_register_types(void)
{
    type_register_static(&bbram_ctrl_info);
}

type_init(bbram_ctrl_register_types)
